/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

package sociogram;

import java.io.PrintStream;
import java.text.DecimalFormat;
import java.util.Iterator;
import java.util.List;

/**
 * It represents a graph with a matrix with the relations among individuals.
 * @author Iván García-Magariño
 */
public class Sociogram {
    /** Number of Individuals */
    private int size;
    
    /** Adyacency matrix with their relationship with values in the 
     * interval [-1,+1], where negative values expresses rejection and positive
     values represents selection **/
    private double matrix [][];
    
     /** Trend matrix that represents the trend of each relatinship 
      * with values in the interval [-1,+1], where negative values expresses 
      * decrease of the reletaionship, and positive represents the increase
      * of the relationship, **/
    private double trend[][];
    
    /** Vector of agents **/
    private IndividualAgent agents[];
    
    /** Incremental Value each time an agent makes a social connection. 
     It representes the percentate (in a number from zero to one) of the 
     * distance to the limit that is increased.**/
    private static double CST_INC=0.15;
    
    // Constants for saving the sociogram to a text file 
    /** Format of the double values printed in the text file for representing
     * the sociogram relations.*/
    private static DecimalFormat RELATION_FORMAT=new DecimalFormat("#.0000");
    /** Format of the double values printed in the text file for representing
     * the sociogrametric values, and it is also used from the GUI.*/
    public static DecimalFormat METRIC_FORMAT=new DecimalFormat("#.00000");
    /** Number of characters that is used for representing the column header**/
    private static int COLUMN_HEADER_LENGTH=23;
    /** Number of characters that is used for representing each column**/
    public static int COLUMN_LENGTH=8;
    
    // Constants for measuring the sociometrics 
    /** Relationship Value from which it is considered to be a selection. 
     * This value must be within the [0, 1] interval.
     */
    private static double SELECTION_THRESHOLD = 0.5;
    /** Relationship Value from which it is considered to be a rejection. 
     * This value must be within the [-1, 0] interval.
     */
    private static double REJECTION_THRESHOLD = -0.5;
    
    
    /** Constructor of the Class. It creates an adjacency matrix from a list
     of agents, relating them with a default value of zero. 
     @param list The list of agents for building the sociogram. **/
    public Sociogram(List<IndividualAgent> list){
        // Establish the number of agents of the sociogram
        size=list.size();
        // Set the vector of agents and establish the index of agents.
        agents=new IndividualAgent[size];
        int index=0;
        IndividualAgent agent;
        Iterator<IndividualAgent> it=list.iterator();
        while(it.hasNext()){
            agent=it.next();
            agent.setIndex(index);
            agents[index]=agent;
            index++;
        }
        
        // Create and set the adjecency matrix and trend matrix
        matrix=new double[size][size];
        trend=new double[size][size];
        for(int i=0;i<size;i++){
            for(int j=0;j<size;j++){
                matrix[i][j]=0;
                trend[i][j]=0;
            }
        }
        
    }
    /** Update a social connection in the sociogram based on a trend
     @param agentA the first agent of the relation
     @param agentB the second agent of the relation
     @param trend the trend of the relation in the interval [-1,1] **/
    public void updateConnection(IndividualAgent agentA, IndividualAgent agentB,
                                double trend){
        // Get Indexes
        int indexA=agentA.getIndex();
        int indexB=agentB.getIndex();
        if(indexA<0 || indexB<0){
            System.out.println("Sociogram:makeConnection. Agents with negative indexes.");
            return;
        }
        // Update the connection betwee agents depending the trend, increasing
        // a porcentage of the distance to the limit
        double relation, inc;
        /*
        relation=matrix[indexA][indexB];
        if(trend>0){
            inc=trend*(1-relation)*CST_INC;
        }else if(trend<0){
            inc=trend*(relation-(-1))*CST_INC;
        }else{
            inc=0;
        }
        matrix[indexA][indexB]=relation+inc;
        */
        // In one single step to reduce the variability according to the
        // recommendations of deck random generators
        matrix[indexA][indexB]=trend;
    }
    
    /** Sets the the trend in a pair of agents 
    @param agentA the first agent of the relation
    @param agentB the second agent of the relation
    @param trend the trend of the relation in the interval [-1,1] **/
    public void setTrend(IndividualAgent agentA, IndividualAgent agentB,
                                double trend){
        // Get Indexes
        int indexA=agentA.getIndex();
        int indexB=agentB.getIndex();
        if(indexA<0 || indexB<0){
            System.out.println("Sociogram:setTrend. Agents with negative indexes.");
            return;
        }
        // Sets the trend
        this.trend[indexA][indexB]=trend;
    }
    
    /** Sets the the trend in a pair of agents 
    @param agentA the first agent of the relation
    @param agentB the second agent of the relation **/
    public double getTrend(IndividualAgent agentA, IndividualAgent agentB){
        // Get Indexes
        int indexA=agentA.getIndex();
        int indexB=agentB.getIndex();
        if(indexA<0 || indexB<0){
            System.out.println("Sociogram:getTrend. Agents with negative indexes.");
            return 0;
        }
        // Sets the trend
        return trend[indexA][indexB];
    }
    /** Saves the sociogram in a text field determined by the stream of the 
     * parameter.
     * @param out the stream where the sociagram content is printed
     */
    public void save(PrintStream out){
        ///// Save the adjacency matrix  
        String str;
        out.println();
        out.println("RESULTING SOCIOGRAM");
        out.println("Adjacency matrix:");
        // Print the header of columns 
        str= " ";
        str=String.format("%1$"+COLUMN_HEADER_LENGTH+"s", str);
        out.print(str+"\t");
        for(int i =0; i<matrix.length; i++){
            str="Agent "+i;
            str=String.format("%1$"+COLUMN_LENGTH+"s", str);
            out.print(str+"\t");
        }
        out.println();
        // Print the adjacency matrix alongside the header of columns
        for(int i =0; i<matrix.length; i++){
            // Print the Header of rows
            String strAgent="Agent "+i;
            String strType=agents[i].getType().toString();
            str=strAgent+"("+strType+")";
            str=String.format("%1$"+COLUMN_HEADER_LENGTH+"s", str);
            out.print(str+"\t");
            // Print each row of the adjacency matrix
            for(int j=0; j<matrix[0].length;j++){
                str=RELATION_FORMAT.format(matrix[i][j]);
                str=String.format("%1$"+COLUMN_LENGTH+"s", str);
                out.print(str+"\t");                
            } 
            out.println();
        }
        ///// Print the measures
        out.println();
        out.println("SOCIOMETRICS OF THE RESULTING SOCIOGRAM");
        out.println("Cohesion:        IAg = "+METRIC_FORMAT.format(measureIAg()));
        out.println("Coherence:       ICg = "+METRIC_FORMAT.format(measureICg()));
        out.println("Dissociation:    IDg = "+METRIC_FORMAT.format(measureIDg()));
        out.println("Group Intensity: IIg = "+METRIC_FORMAT.format(measureIIg()));
    }
    
    // GROUP SOCIOMETRICS IN THE SOCIOGRAM
    /* 
    References:
    Barrasa, A. Gil, F. (2004). Un programa informático para el 
    cálculo y la representación de índices y valores sociométricos. 
    Psicothema, 16(2), 329-335.
    
    /** Measure the Cohesion of the group with the sociometric IAg described in 
    (Barrasa & Gil, 2008).
    **/
    public double measureIAg(){
        // Variables
        double n, Rei, sum, IAg;
        // Get the size of the group
        n=agents.length;
        // Calculate the sum of Re for each agent i
        sum=0;
        for(int i=0; i<agents.length; i++){
            Rei=measureRe(i);
            sum+=Rei;
        }
        // Divide the sum for obtainint IAg 
        IAg=sum/(n*(n-1));
        return IAg;
    }
    
     /** Measure the Dissociation of the group with the sociometric IDg described in 
    (Barrasa & Gil, 2008).
    **/
    public double measureIDg(){
        // Variables
        double n, Rri, sum, IAg;
        // Get the size of the group
        n=agents.length;
        // Calculate the sum of Rr for each agent i
        sum=0;
        for(int i=0; i<agents.length; i++){
            Rri=measureRr(i);
            sum+=Rri;
        }
        // Divide the sum for obtainint IDg 
        IAg=sum/(n*(n-1));
        return IAg;
    }
    
     /** Measure the Coherence of the group with the sociometric ICg described in 
    (Barrasa & Gil, 2008). For practical purposes, it returns zero when there
    * ist not any selction in the sociogram.
    **/
    public double measureICg(){
        // Variables
        double n, Rei, Sei, sumRei, sumSei, metric;
        // Get the size of the group
        n=agents.length;
        // Calculate the sum of Re and the sum of  Se
        sumRei=0;
        sumSei=0;
        for(int i=0; i<agents.length; i++){
            sumRei+=measureRe(i);
            sumSei+=measureSe(i);
        }
        // If denominator greater than zero, perfor the division to calculate
        // the metric
        if(sumSei>0){
            metric=sumRei/sumSei;
        }else{
            metric=0;
        }
        return metric;
    }
    
     /** Measure the Intensity of the Group with the sociometric IIg described 
      * in (Barrasa & Gil, 2008).
    **/
    public double measureIIg(){
        // Variables
        double n, sumRei, sumSri, metric;
        // Get the size of the group
        n=agents.length;
        // Calculate the sum of Re and the sum of  Se
        sumRei=0;
        sumSri=0;
        for(int i=0; i<agents.length; i++){
            sumRei+=measureRe(i);
            sumSri+=measureSr(i);
        }
        // Calculate the sociometric
        metric=(sumRei+sumSri)/(n-1);
        return metric;
    }
    
    // INDIVIDUAL SOCIOMETRICS IN THE SOCIOGRAM
    /** Measure the Reciprocity of Selections: number of selections of an 
     * agent that have been selected back with reciprocal selections.
     * @param indexAgent the index of the agent that is measured
     * @return the value of the sociometric
     */
    public int measureRe(int indexAgent){
        int Re;
        double relation, reciprocal;
        // Check whether the indes is within the limits
        if(indexAgent<0 || indexAgent>=agents.length){
            System.out.println("Agent index out of bounds when measuring Re...");
            return -1;
        }
        // Count the metric
        Re=0;
        for (int j=0; j<agents.length; j++){
            relation=matrix[indexAgent][j];
            reciprocal=matrix[j][indexAgent];
            if(relation>SELECTION_THRESHOLD && 
                    reciprocal>SELECTION_THRESHOLD){
                Re++;
            }
        }
        return Re;
    }
    /** Measure the Reciprocity of Rejections: number of rejections of an 
     * agent that have been rejected back with reciprocal rejections.
     * @param indexAgent the index of the agent that is measured
     * @return the value of the sociometric
     */
    public int measureRr(int indexAgent){
        int Rr;
        double relation, reciprocal;
        // Check whether the indes is within the limits
        if(indexAgent<0 || indexAgent>=agents.length){
            System.out.println("Agent index out of bounds when measuring Rr...");
            return -1;
        }
        // Count the metric
        Rr=0;
        for (int j=0; j<agents.length; j++){
            relation=matrix[indexAgent][j];
            reciprocal=matrix[j][indexAgent];
            if(relation<REJECTION_THRESHOLD && 
                    reciprocal<REJECTION_THRESHOLD){
                Rr++;
            }
        }
        return Rr;
    }
    
    /** Measure the selections receibed by subject.
     * @param indexAgent the index of the agent that is measured
     * @return the value of the sociometric
     */
    public int measureSe(int indexAgent){
        int metric;
        double reciprocal;
        // Check whether the indes is within the limits
        if(indexAgent<0 || indexAgent>=agents.length){
            System.out.println("Agent index out of bounds when measuring Se...");
            return -1;
        }
        // Count the metric
        metric=0;
        for (int j=0; j<agents.length; j++){
            reciprocal=matrix[j][indexAgent];
            if(reciprocal>SELECTION_THRESHOLD){
                metric++;
            }
        }
        return metric;
    }
    /** Measure the rejections receibed by the subject.
     * @param indexAgent the index of the agent that is measured
     * @return the value of the sociometric
     */
    public int measureSr(int indexAgent){
        int metric;
        double reciprocal;
        // Check whether the indes is within the limits
        if(indexAgent<0 || indexAgent>=agents.length){
            System.out.println("Agent index out of bounds when measuring Se...");
            return -1;
        }
        // Count the metric
        metric=0;
        for (int j=0; j<agents.length; j++){
            reciprocal=matrix[j][indexAgent];
            if(reciprocal<REJECTION_THRESHOLD){
                metric++;
            }
        }
        return metric;
    }
    

    // GETTERS AND SETTERS 
    public int getSize() {
        return size;
    }

    public IndividualAgent[] getAgents() {
        return agents;
    }

    public double[][] getMatrix() {
        return matrix;
    }  
}
